//------------------------------------------------------------------------------------------------------
//
//  HappyTerminal.c
//
// ToothPIC Terminal Emulator application 
//
//  (c) FlexiPanel Ltd
//
//-------------------------------------------------------------------------------------------------------

#include "Toothpick.h"
#include <p18cxxx.h>
#include "HappyTerminalRes.h"

// HappyTerminal semaphores
unsigned char HTSemaphores = 0;
#define HTS_BufferOn 				0x01	// Comparators configured as buffers
#define HTS_InvertOn 				0x02	// Comparators configured as inverters
#define HTS_FCON 					0x04	// flow control on
#define HTS_FCInv 					0x08  	// flow control pins are inverted
#define HTS_ErrorLastCycle			0x10	// UART error happened 
#define HTS_NotMainDialog			0x20	// main dialog is not visible
#define HTS_ReceiveOptionsChanged	0x40  	// must hide / show debug text controls
#define HTS_ResetMePlease			0x80  	// new for V3.0.00005 - reset BlueMatik on general error

// Receive buffer
#define HTRxBuffLen 256
#pragma udata HTRxBuff_Section
unsigned unsigned char HTRxBuff[ HTRxBuffLen ];
#pragma udata
volatile unsigned char HTRxStart;			// Receive buffer startpoint
volatile unsigned char HTRxNext;		// Receive buffer endpoint
#define HTRxNChar 			( (unsigned char) (HTRxNext - HTRxStart ) )
#define HTRxSpace			( (unsigned char) ( 0xFF - HTRxNChar ) )
#define HTRxEmpty 			( HTRxNext == HTRxStart )
#define HTClear				((HTSemaphores & HTS_FCInv) ? 1 : 0)
#define HTStop				((HTSemaphores & HTS_FCInv) ? 0 : 1)
#define FC_MAX  220 // number of chars in buffer before flow control triggered

#define HTRxLoc( i )		( HTRxBuff + ( HTRxStart + ( (unsigned char) i ) ) )
#define HTRxCh				( HTRxBuff + HTRxStart )

// terminal screen control
unsigned char BottomLineEndLoc = 0;  // position of cursor on bottom line
rom unsigned char szError[] = "Error";

// flow control pins
#define RTSPin	CCP5Pin
#define DirRTS	DirCCP5
#define CTSPin	CCP4Pin
#define DirCTS	DirCCP4

// local function prototypes
void SetupUART( void );
void ShowHideDbg(unsigned short Ctl );
void CopyTextLine( unsigned char * pDest, unsigned char * pSrc );
void CopyASCLine( unsigned char * pDest, unsigned char * pSrc );
void StartNewLine( void );
void UpdateCtrl( unsigned short ID );
void ShowHideASCII( void );
void ClearBottomLine( void );
void TxChar( unsigned char Chr, unsigned char bEcho );

void main( void )
{

	// if no LinkMatik, flash red led rapidly
	while ((ToothpickSemaphores&TPSF_LMTEXISTS)==0)
	{
		LedRed = ~LedRed;
		msDelay(50);
	}

	// start user interface
	FxPCommand( FxPC_Start, 0, 0 );		// Start FlexiPanel service
	AwaitLMTComplete();					// Service operating, no client
	ClearBottomLine();					// filles entire bottom line with nulls

	// initialize UART
	HTRxStart = 0;
	HTRxNext = 0;
	InitUART( SPBaudBrghOne9600, 1 );
	SetupUART();

	// main loop
	while ( 1 ) 
	{
		// Flash LED to show we're alive; short puls for not connected,
		// long pulse for connected
		unsigned char iCount;
		LedGreen = 1;
		for (iCount = 0; iCount< 5; iCount++)
		{
			msDelay( 100 );
			if (iCount == ( NumConnections ? 3:0) ) LedGreen = 0;
		}

		// processs Rx errors
		if (UARTEnable == 0)
		{
			// an error must have occurred
			if ((HTSemaphores & HTS_ErrorLastCycle) == 0)
			{
				// Error occurred this time through the loop; wait one cycle so Red LED is clearly visible
				HTSemaphores |= HTS_ErrorLastCycle;
			}
			else
			{
				// Error occured during previous loop; can reset UART Rx now
				HTSemaphores &= ~HTS_ErrorLastCycle;
				LedRed = LedRedOff;
				UARTEnable = 1;
			}
		}

		// new for V3.0.00005 - reset BlueMatik on general error
		if (HTSemaphores & HTS_ResetMePlease)
		{
			Reset();						
		}

	}  // end of while ( 1 ) 

}

// HighInterrupt handler 
void HighInterrupt (void)
{
	if (UARTRxIntFlag)
	{
		// note PIR3bits.TX2IF cleared by hardware
		if (UARTOverrunErr || UARTFrameErr) 
		{
			unsigned char ch;
			// turn on red led to indicate an error
			LedRed = LedRedOn;
			UARTEnable = 0;
			ch = UARTChIn;  // clears UARTRxIntFlag
			return;
		}

		// store character and advance buffer
		*( HTRxBuff + HTRxNext ) = UARTChIn;
		HTRxNext++;

		// If flow control, tell RTS to stop if getting near full
		if ( HTSemaphores & HTS_FCON && LMTRxNChar > FC_MAX )  	RTSPin = HTStop;

		// raise software interrupt to process character
		SetSWI( SWI_SWI1 );

		return;
	}
}


// LowInterrupt handler 
void LowInterrupt (void)
{
	if (IsSWI( SWI_Tick ) )
	{
		ClearSWI( SWI_Tick );		// Clear clock interrupt flag
		return;
	}

	if (IsSWI( SWI_SWI1 ) )
	{
		// store local copies of RxStart and RxNext in case they change during this routine
		unsigned char tmpHTRxNext = HTRxNext;
		unsigned char tmpHTRxStart = HTRxStart;
		unsigned char pCh;

		// debug mode or ascii ?
		unsigned long lDebugMode;
		Get_ReceiveOptions_2B( &lDebugMode );

		// process character
		for (pCh = tmpHTRxStart; pCh != tmpHTRxNext; pCh++)
		{
			// get character
			unsigned char chr = HTRxBuff[pCh];

			if (lDebugMode)
			{	// debug mode
				unsigned char HexOffest = BottomLineEndLoc*3;
				
				// add hex text to pText19_6E		
				pText19_6E[	HexOffest ] = (chr >> 4) + '0';
				if (pText19_6E[	HexOffest ] > '9') pText19_6E[	HexOffest ] += 0x07;
				HexOffest++;
				pText19_6E[	HexOffest ] = (chr & 0x0F) + '0';
				if (pText19_6E[	HexOffest ] > '9') pText19_6E[	HexOffest ] += 0x07;
				if (BottomLineEndLoc<MaxBytes_ASC19_5B-2)
				{
					HexOffest++;
					pText19_6E[	HexOffest ] = '.';
				}
				HexOffest++;
				pText19_6E[	HexOffest ] = 0;

				// add ascii to pASC19_5B, use underline for ctrl chr
				if (chr<0x20 || chr>0x7E) chr = '_'; 
				pASC19_5B[BottomLineEndLoc] = chr;
				
				// incrememnt and startt a new line if necessary
				BottomLineEndLoc ++;
				if (BottomLineEndLoc == MaxBytes_ASC19_5B-1)
				{
					StartNewLine();
				}
			}
			else
			{
				// ASCII - strictly ASCII, so drop bit 7
				chr = chr & 0x7F;

				// DEL character
				if (chr == 0x7F) // DEL
				{
					if (BottomLineEndLoc>0)
					{
						BottomLineEndLoc--;
					}
				}
				// is control character ?
				else if (chr < 0x20)
				{
					if (chr==0x0D)  // CR
					{
						unsigned char bAppendLF;
						BottomLineEndLoc = 0;
	
						Get_Add_line_feed_2C( &bAppendLF );
						if (bAppendLF)  StartNewLine();
					}
					else if (chr==0x0A)  // LF
					{
						StartNewLine();
					}
				}
				else
				{
					if (BottomLineEndLoc<MaxBytes_Text19_6E-1)
					{
						pText19_6E[BottomLineEndLoc] = chr;
						BottomLineEndLoc ++;
					}
					// wrap if required
					if (BottomLineEndLoc==MaxBytes_Text19_6E-1)
					{
						unsigned char bWrapRx;
						Get_Wrap_lines_2D( &bWrapRx );
						if (bWrapRx)
						{
							StartNewLine();
						}
					}
				}
			}
		}
        HTRxStart = tmpHTRxNext; 
		if ( HTRxNChar < 250 )  RTSPin = HTClear; 

		FxPCommand( FxPC_CtlUpdate, ID_Text19_6E, 0 );
		if (lDebugMode) FxPCommand( FxPC_CtlUpdate, ID_ASC19_5B, 0 );

		if (HTRxNext==HTRxStart) ClearSWI( SWI_SWI1 );		// Clear SWI flag
		return;
	}

}


// Error event handler 
void ErrorStatus( unsigned char ErrNum )
{
	// Flash error number for diagnositic purposes
	// For product releases, a Reset() is better for unanticipated errors
	Breakpoint( ErrNum );
}


// BlueMatik Event handler 
void LMTEvent( unsigned char EventID, void *pData1, void *pData2 )
{
	if ( EventID==LMTE_Syntax )
	{
		if (((unsigned char*)pData1)[0] == 0x0F) // general error
		{
			HTSemaphores |= HTS_ResetMePlease;
			return;
		}

		// Flash error number for diagnositic purposes
		// For product releases, a Reset() is better for unanticipated errors
		Breakpoint( ((unsigned char*)pData1)[0] );
	}
}

// FxPEvent handler 
void FxPEvent( unsigned char EventID, void *pData )
{
	if ( EventID==FxPE_Disco )
	{ 
		// reset active dialog to main dialog 
		FxPCommand( FxPC_SetDialog, DlgID_ASC0_15, 0 );	
		HTSemaphores &= ~HTS_NotMainDialog;
		return;
	}

	if ( EventID==FxPE_ClntUpdate ) 
	{
		// close buttons
		if (*((unsigned short*) pData) == ID_Close_22 || *((unsigned short*) pData) == ID_Close_2F)
		{
			// close button presed - disconnect
			FxPCommand( FxPC_Disconnect, 0, 0 );	
			// turn off red led after disconnection 
			LedRed = LedRedOff;
		}

		// main dialog settings button
		else if (*((unsigned short*) pData) == ID_Settings_21)
		{
			// switch to settings dialog
			FxPCommand( FxPC_SetDialog, DlgID_BaudTitle_37, 0 );	
			HTSemaphores |= HTS_NotMainDialog;
		}

		// settings dialog OK button
		else if (*((unsigned short*) pData) == ID_OK_2E)
		{
			// note any changes
			SetupUART();

			// start new line if Receive Options Changed
			if ( HTSemaphores & HTS_ReceiveOptionsChanged )
			{
				StartNewLine();
			}

			// switch to settings dialog
			FxPCommand( FxPC_SetDialog, DlgID_ASC0_15, 0 );	
			HTSemaphores &= ~(HTS_NotMainDialog+HTS_ReceiveOptionsChanged);
		}

		// settings dialog Receive options changed
		else if (*((unsigned short*) pData) == ID_ReceiveOptions_2B)
		{
			HTSemaphores |= HTS_ReceiveOptionsChanged;
		}

		// transmit text OK button pressed
		else if (*((unsigned short*) pData) == ID_OutCharsOK_20)
		{
			unsigned long bHex;
			unsigned char bEcho;
			unsigned char bCR;
			unsigned char bLF;
			unsigned char ChLoc;
			unsigned char bError = 0;

			Get_TransmitOptions_26( &bHex );
			Get_Add_return_28( &bCR );
			Get_Add_line_feed_6F( &bLF );
			Get_Echo_29( &bEcho );

			for (ChLoc = 0; pOutCharsOK_20[ChLoc]!=0  && ! bError; ChLoc++)
			{
				unsigned char byte;
				// get byte
				if (bHex)
				{
					unsigned char tmpByte;
					byte = pOutCharsOK_20[ChLoc] - '0';
					if (byte>0x09) byte -= 0x07;  // convert A-F
					if (byte>0x0F) byte -= 0x20;  // convert lower case A-F
					if (byte>0x0F) 
					{
						bError = 0xFF;
						break;	// error, not hex
					}
					ChLoc++;
					if (pOutCharsOK_20[ChLoc] == 0)  
					{
						bError = 0xFF;
						break ; // Error - two digits required
					}
										
					tmpByte = pOutCharsOK_20[ChLoc] - '0';
					if (tmpByte>0x09) tmpByte -= 0x07;  // convert A-F
					if (tmpByte>0x0F) tmpByte -= 0x20;  // convert lower case A-F
					if (tmpByte>0x0F) 
					{
						bError = 0xFF;
						break;	// error, not hex
					}
					byte = (byte<<4) + tmpByte;
				}
				else
				{
					byte= pOutCharsOK_20[ChLoc];
				}
				// transmit character
				TxChar( byte, bEcho );
			}

			if (bCR) TxChar( '\r', bEcho );
			if (bLF) TxChar( '\n', bEcho );

			// clear edit text
			if (bError)
			{
				Set_OutCharsOK_20( szError, 0 );
			}
			else
			{
				pOutCharsOK_20[0] = 0;
			}
			FxPCommand( FxPC_CtlUpdate, ID_OutCharsOK_20, 0 );
		}

		// new baud target value 
		else if (*((unsigned short*) pData) == ID_Baud_Target_34)
		{
			// update actual baud value
			SetupUART();
		}
		return;
	}

	if ( EventID==FxPE_NewDialog ) 
	{
		// visibility / invisibility of ASCII controls is stored on client only, not server,
		// so must retransmit correct state once dialog is redisplayed
		if ( ((unsigned short)pData) == DlgID_ASC0_15) ShowHideASCII();
	}

}

// set up uart on the basis of the settings dialog
void SetupUART( void )
{
	unsigned long lval ;  
	unsigned char chval;

	// calculate closest baud rate achievable 
	Get_Baud_Target_34( &lval );

	// set baud rate
	if (lval < 1221)
	{
		SPBRG2 = 255;
		TXSTA2bits.BRGH = 0;
		* plBaud_Actual_35 = 1221;
	}
	else if (lval < 4884)
	{
		unsigned long BestVal = ((24000000 / 32 / lval ) - 1) / 2;
		TXSTA2bits.BRGH = 0;
		SPBRG2 = (unsigned char) BestVal;
		* plBaud_Actual_35 = 24000000 / 64 / (((unsigned long)SPBRG2)+ 1);
	}
	else if (lval < 115200)  // maximum since high interrupts can't respond faster than this
	{
		unsigned long BestVal = ((24000000 / 8 / lval ) - 1) / 2;
		TXSTA2bits.BRGH = 1;
		SPBRG2 = (unsigned char) BestVal;
		* plBaud_Actual_35 = 24000000 / 16 / (((unsigned long)SPBRG2)+ 1);
	}
	else 
	{
		TXSTA2bits.BRGH = 1;
		SPBRG2 = 10;
		* plBaud_Actual_35 = 113630;
	}

	FxPCommand( FxPC_CtlUpdate, ID_Baud_Actual_35, 0 );

	// note in bitfield whether inverter / buffers used or not
	HTSemaphores &= ~( HTS_BufferOn | HTS_InvertOn | HTS_FCON | HTS_FCInv );
	Get_Buffers_70( &lval );
	if (lval==1) HTSemaphores |= HTS_BufferOn;
	if (lval==2) HTSemaphores |= HTS_InvertOn;
	Get_Flow_Control_36( &lval );
	if (lval&0x01) HTSemaphores |= HTS_FCON;
	if (lval&0x02) HTSemaphores |= HTS_FCON + HTS_FCInv;

	// initialize flow control, rts clear 
	if ( HTSemaphores & HTS_FCON)
	{
		RTSPin = HTClear;
		DirRTS = 0;  // rts clear output
		DirCTS = 1;  // cts input
	}

	// set up comparators as inverters if required
	if (HTSemaphores & HTS_BufferOn)
	{
		CmpsBuffer;
	}
	else if (HTSemaphores & HTS_InvertOn)
	{
		CmpsInvert;
	}
	else
	{
		CmpsOff;
	}

}

void ShowHideASCII( void )
{
	FxPCommand( FxPC_MultiPropsUpdate, 20, 0 ); 

	ShowHideDbg(ID_ASC0_15);
	ShowHideDbg(ID_ASC1_49);
	ShowHideDbg(ID_ASC2_4A);
	ShowHideDbg(ID_ASC3_4B);
	ShowHideDbg(ID_ASC4_4C);
	ShowHideDbg(ID_ASC5_4D);
	ShowHideDbg(ID_ASC6_4E);
	ShowHideDbg(ID_ASC7_4F);
	ShowHideDbg(ID_ASC8_50);
	ShowHideDbg(ID_ASC9_51);
	ShowHideDbg(ID_ASC10_52);
	ShowHideDbg(ID_ASC11_53);
	ShowHideDbg(ID_ASC12_54);
	ShowHideDbg(ID_ASC13_55);
	ShowHideDbg(ID_ASC14_56);
	ShowHideDbg(ID_ASC15_57);
	ShowHideDbg(ID_ASC16_58);
	ShowHideDbg(ID_ASC17_59);
	ShowHideDbg(ID_ASC18_5A);
	ShowHideDbg(ID_ASC19_5B);
}

void ShowHideDbg(unsigned short Ctl )
{ 
	unsigned long lVal;
	Get_ReceiveOptions_2B( &lVal );
	if (lVal) 
		lVal = 0x00000000;
	else
		lVal = 0x00000001;

	FxPCommand( FxPC_SetProps, Ctl, &lVal ); 
}


void StartNewLine( void )
{
	// scroll up one line
	CopyTextLine( pText0_1, pText1_5C );
	CopyTextLine( pText1_5C, pText2_5D );
	CopyTextLine( pText2_5D, pText3_5E );
	CopyTextLine( pText3_5E, pText4_5F );
	CopyTextLine( pText4_5F, pText5_60 );
	CopyTextLine( pText5_60, pText6_61 );
	CopyTextLine( pText6_61, pText7_62 );
	CopyTextLine( pText7_62, pText8_63 );
	CopyTextLine( pText8_63, pText9_64 );
	CopyTextLine( pText9_64, pText10_65 );
	CopyTextLine( pText10_65, pText11_66 );
	CopyTextLine( pText11_66, pText12_67 );
	CopyTextLine( pText12_67, pText13_68 );
	CopyTextLine( pText13_68, pText14_69 );
	CopyTextLine( pText14_69, pText15_6A );
	CopyTextLine( pText15_6A, pText16_6B );
	CopyTextLine( pText16_6B, pText17_6C );
	CopyTextLine( pText17_6C, pText18_6D );
	CopyTextLine( pText18_6D, pText19_6E );
	CopyASCLine( pASC0_15, pASC1_49 );
	CopyASCLine( pASC1_49, pASC2_4A );
	CopyASCLine( pASC2_4A, pASC3_4B );
	CopyASCLine( pASC3_4B, pASC4_4C );
	CopyASCLine( pASC4_4C, pASC5_4D );
	CopyASCLine( pASC5_4D, pASC6_4E );
	CopyASCLine( pASC6_4E, pASC7_4F );
	CopyASCLine( pASC7_4F, pASC8_50 );
	CopyASCLine( pASC8_50, pASC9_51 );
	CopyASCLine( pASC9_51, pASC10_52 );
	CopyASCLine( pASC10_52, pASC11_53 );
	CopyASCLine( pASC11_53, pASC12_54 );
	CopyASCLine( pASC12_54, pASC13_55 );
	CopyASCLine( pASC13_55, pASC14_56 );
	CopyASCLine( pASC14_56, pASC15_57 );
	CopyASCLine( pASC15_57, pASC16_58 );
	CopyASCLine( pASC16_58, pASC17_59 );
	CopyASCLine( pASC17_59, pASC18_5A );
	CopyASCLine( pASC18_5A, pASC19_5B );

	// initialize new line
	BottomLineEndLoc = 0;
	ClearBottomLine();

	// if visible, update screen
	if ((HTSemaphores & HTS_NotMainDialog)==0)
	{
		FxPCommand( FxPC_MultiUpdate, 40, 0 ); 
		UpdateCtrl( ID_ASC0_15 );
		UpdateCtrl( ID_Text0_1 );
		UpdateCtrl( ID_ASC1_49 );
		UpdateCtrl( ID_Text1_5C );
		UpdateCtrl( ID_ASC2_4A );
		UpdateCtrl( ID_Text2_5D );
		UpdateCtrl( ID_ASC3_4B );
		UpdateCtrl( ID_Text3_5E );
		UpdateCtrl( ID_ASC4_4C );
		UpdateCtrl( ID_Text4_5F );
		UpdateCtrl( ID_ASC5_4D );
		UpdateCtrl( ID_Text5_60 );
		UpdateCtrl( ID_ASC6_4E );
		UpdateCtrl( ID_Text6_61 );
		UpdateCtrl( ID_ASC7_4F );
		UpdateCtrl( ID_Text7_62 );
		UpdateCtrl( ID_Text8_63 );
		UpdateCtrl( ID_ASC8_50 );
		UpdateCtrl( ID_Text9_64 );
		UpdateCtrl( ID_ASC9_51 );
		UpdateCtrl( ID_Text10_65 );
		UpdateCtrl( ID_ASC10_52 );
		UpdateCtrl( ID_Text11_66 );
		UpdateCtrl( ID_ASC11_53 );
		UpdateCtrl( ID_Text12_67 );
		UpdateCtrl( ID_ASC12_54 );
		UpdateCtrl( ID_Text13_68 );
		UpdateCtrl( ID_ASC13_55 );
		UpdateCtrl( ID_Text14_69 );
		UpdateCtrl( ID_ASC14_56 );
		UpdateCtrl( ID_Text15_6A );
		UpdateCtrl( ID_ASC15_57 );
		UpdateCtrl( ID_Text16_6B );
		UpdateCtrl( ID_ASC16_58 );
		UpdateCtrl( ID_Text17_6C );
		UpdateCtrl( ID_ASC17_59 );
		UpdateCtrl( ID_Text18_6D );
		UpdateCtrl( ID_ASC18_5A );
		UpdateCtrl( ID_Text19_6E );
		UpdateCtrl( ID_ASC19_5B );
	}
}

void ClearBottomLine( void )
{
	unsigned char ch;
	for (ch = 0; ch<MaxBytes_Text19_6E; ch++)
	{
		pText19_6E[ch] = 0;
	}
	for (ch = 0; ch<MaxBytes_ASC19_5B; ch++)
	{
		pASC19_5B[ch] = 0;
	}
}

void CopyTextLine( unsigned char * pDest, unsigned char * pSrc )
{
	unsigned char ch;
	for (ch = MaxBytes_Text19_6E; ch>0; ch--)
	{
		*pDest = *pSrc;
		pDest++;
		pSrc++;
	}
} 

void CopyASCLine( unsigned char * pDest, unsigned char *pSrc )
{
	unsigned char ch;
	for (ch = MaxBytes_ASC19_5B; ch>0; ch--)
	{
		*pDest = *pSrc;
		pDest++;
		pSrc++;
	}
} 

void UpdateCtrl( unsigned short ID )
{
	FxPCommand( FxPC_CtlUpdate, ID, 0 );
}


void TxChar( unsigned char Chr, unsigned char bEcho )
{
	// wait for TxReg to be clear for loading
	while (UARTNotTxmitting==0) {;};

	// if echo, append to input buffer
	if (bEcho && HTRxSpace > 2)
	{
		// store character and advance buffer
		DisableHighInts;
		*( HTRxBuff + HTRxNext ) = Chr;
		HTRxNext++;
		EnableHighInts;

		// If flow control, tell RTS to stop if getting near full
		if ( HTSemaphores & HTS_FCON && LMTRxNChar > FC_MAX )  	RTSPin = HTStop;

		// raise software interrupt to process character
		SetSWI( SWI_SWI1 );
	}

	// wait for CTS
	if ( HTSemaphores & HTS_FCON ) 
	{
		while ( CTSPin == HTStop ) {;};
	}
	
	// transmit char
	UARTChOut = Chr;
}
